/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * encode.c: encoding loop
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/socket.h>
#include <speex/speex.h>

#include "speextalk.h"

#define BUF_SIZE 1024

static unsigned sequence_out = 0;
static int bandwidth_out_acc = 0;
static unsigned packet_timestamp_out = 0;

/*
 * Prepares the recording channel according to the selected audio format.
 */
static void
prepare_recording_with_af(void)
{
    const audio_format_t *af;

    check_audio_format(audio_format_out, "");
    af = &audio_formats[audio_format_out - 1];
    prepare_recording(af->sample_rate, af->channels);
}

/*
 * Prepares the header of an outgoing packet.
 */
static void
prepare_packet(unsigned char *p, int *pl)
{
    p[0] = 1;
    p[1] = audio_format_out;
    p[2] = (sequence_out >> 8) & 0xFF;
    p[3] = sequence_out & 0xFF;
    *pl = 4;
}

/*
 * Sends a packet. Dies on fatal error.
 */
static void
send_packet(unsigned char *p, int pl)
{
    while(send(sock, p, pl, 0) < 0) {
	if(errno == ECONNREFUSED) {
	    pthread_mutex_lock(&global_mutex);
	    running = 0;
	    pthread_mutex_unlock(&global_mutex);
	    pthread_exit(NULL);
	}
	if(errno != EINTR)
	    fatal_system("send");
    }
    update_bandwidth(&bandwidth_out, &bandwidth_out_acc,
	&packet_timestamp_out, pl);
}

/*
 * Repeatedly read some sound, encodes it and sends it.
 */
void *
encode_loop(void *dummy)
{
    (void)dummy;
    SpeexBits bits;
    void *enc_state;
    int frame_size;
    short buf[BUF_SIZE];
    int buf_p = 0;
    unsigned char packet[max_packet_size];
    int packet_pos = 0;
    int packet_frames = 0;
    int n;
    int run;

    prepare_recording_with_af();
    speex_bits_init(&bits);
    enc_state =
	speex_encoder_init(audio_formats[audio_format_out - 1].speex_mode);
    speex_encoder_ctl(enc_state, SPEEX_GET_FRAME_SIZE, &frame_size);
    speex_encoder_ctl(enc_state, SPEEX_SET_QUALITY, &speex_quality);
    speex_encoder_ctl(enc_state, SPEEX_SET_VBR, &speex_vbr);
    while(1) {
	pthread_mutex_lock(&global_mutex);
	run = running;
	pthread_mutex_unlock(&global_mutex);
	if(!run)
	    break;
	if(buf_p < frame_size) {
	    n = read_samples(buf + buf_p, BUF_SIZE - buf_p);
	    if(igain != 100) {
		int i, v;

		for(i = 0; i < n; i++) {
		    v = (int)buf[buf_p + i] * igain / 100;
		    buf[buf_p + i] =
			v < -32767 ? -32767 : v > 32767 ? 32767 : v;
		}
	    }
	    buf_p += n;
	} else {
	    if(packet_pos == 0)
		prepare_packet(packet, &packet_pos);
	    speex_bits_reset(&bits);
	    update_power(&power_out, buf, frame_size);
	    speex_encode_int(enc_state, buf, &bits);
	    n = speex_bits_nbytes(&bits);
	    if(packet_pos + 2 + n > max_packet_size) {
		if(packet_frames == 0) {
		    fprintf(stderr, "Packet overflow\n");
		    exit(1);
		}
		send_packet(packet, packet_pos);
		packet_frames = 0;
		packet_pos = 0;
		prepare_packet(packet, &packet_pos);
	    }
	    packet[packet_pos++] = n >> 8;
	    packet[packet_pos++] = n & 0xFF;
	    speex_bits_write(&bits, packet + packet_pos,
		max_packet_size - packet_pos);
	    buf_p -= frame_size;
	    memmove(buf, buf + frame_size, buf_p * 2);
	    packet_pos += n;
	    packet_frames++;
	    sequence_out++;
	    if(packet_frames >= max_packet_frames) {
		send_packet(packet, packet_pos);
		packet_frames = 0;
		packet_pos = 0;
	    }
	}
    }
    return(NULL);
}
