/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * interface_text.h: simple interface in text mode
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <locale.h>
#include <poll.h>
#include <sys/time.h>
#include <termios.h>
#include <unistd.h>

#include "speextalk.h"

#define HIST_SIZE 8

static struct termios orig_termios;
static int igain_mute = 0;

/*
 * Gets an UTF-8-encoded character from the string pointed by pp, returns
 * the code and updates pp. Returns -1 if pp would overflow e.
 */
static int
get_utf8(const unsigned char **pp, const unsigned char *e)
{
    const unsigned char *p = *pp;
    int ml = e == NULL ? 6 : e - p;
    int i, l, c;

    if(ml == 0)
	return(-1);
    c = *(p++);
    if(c >= 0x80) {
	for(l = 0; l < 7 && (c & (0x80 >> l)) != 0; l++);
	if(l < 2 || l > 6)
	    return(-1);
	c &= 0x7F >> l;
	for(i = 1; i < l; i++)
	    c = (c << 6) | (0x3F & *(p++));
    }
    *pp = p;
    return(c);
}

/*
 * Restores terminal state.
 */
static void
restore_termios(void)
{
    tcsetattr(0, TCSANOW, &orig_termios);
}

/*
 * Sets up the terminal.
 */
void
interface_init(void)
{
    struct termios term;

    if(setlocale(LC_CTYPE, "") == NULL)
	perror("setlocale");
    if(tcgetattr(0, &term))
	return;
    orig_termios = term;
    term.c_lflag &= ~(ECHO | ICANON | IEXTEN);
    term.c_cc[VMIN] = 1;
    term.c_cc[VTIME] = 0;
    atexit(restore_termios);
    tcsetattr(0, TCSANOW, &term);
}

/*
 * Draws in p an histogram for the value v.
 */
static void
draw_hist(unsigned char *p, int v)
{
    int i;

    v = v * HIST_SIZE * HIST_SIZE / 32768;
    for(i = 1; i <= HIST_SIZE; i++)
	*(p++) = i * i <= v ? '#' : '-';
}

/*
 * Reads one byte from stdin.
 */
static int
read_one_char(void)
{
    unsigned char c;
    int l;

    while((l = read(0, &c, 1)) < 0)
	if(errno != EINTR)
	    fatal_system("read");
    return(l == 1 ? c : -1);
}

/*
 * Understands one key and acts.
 */
static void
process_input(void)
{
    unsigned char c;

    c = read_one_char();
    switch(c) {
	case 'q': case 'Q':
	    exit(0);
	case 'm':
	    {
		int t = igain;
		igain = igain_mute;
		igain_mute = t;
	    }
	    break;
    }
}

/*
 * Returns the current timestamp as milliseconds. Wraps every 50- days (if
 * int is 32 bits) but it does not matter since it is only used for relative
 * timestamping over a few seconds.
 */
unsigned
get_now(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return((unsigned)tv.tv_sec * 1000 + (unsigned)tv.tv_usec / 1000);
}

/*
 * Runns the interface until signaled otherwise by the user of the running
 * global.
 */
void
interface_run(void)
{
    unsigned char st[80];
    int i;
    struct pollfd fd = { .fd = 0, .events = POLLIN };
    unsigned next = 0, now;
    int bw_out, bw_in, pw_out, pw_in, run;

    next = get_now() - 1;
    while(1) {
	pthread_mutex_lock(&global_mutex);
	bw_out = bandwidth_out;
	bw_in = bandwidth_in;
	pw_out = power_out;
	pw_in = power_in;
	run = running;
	pthread_mutex_unlock(&global_mutex);
	if(!run)
	    break;
	poll(&fd, 1, 50);
	if((fd.revents & POLLIN) != 0)
	    process_input();
	now = get_now();
	if(now - next < 0x80000000) {
	    next = now + 1000;
	    sprintf(st, "Out: %5d o/s |", bw_out);
	    draw_hist(st + 16, pw_out);
	    for(i = 24; i < 28; i++)
		st[i] = ' ';
	    sprintf(st + 28, "In: %5d o/s |", bw_in);
	    draw_hist(st + 43, pw_in);
	    for(i = 51; i < 55; i++)
		st[i] = ' ';
	    st[55] = igain_mute == 0 ? ' ' : 'M';
	    st[56] = '\r';
	    write(1, st, 57);
	}
    }
}

/*
 * Prints a message to the user. The message is in UTF-8.
 */
void
interface_print(const char *t)
{
    unsigned char b[MB_CUR_MAX];
    int l, c;
    const unsigned char *p = t, *e = strchr(t, 0);

    while(p < e) {
	c = get_utf8(&p, e);
	if(c < 0) {
	    c = '?';
	    t++;
	}
	l = wctomb(b, c);
	if(l <= 0) {
	    b[0] = '?';
	    l = 1;
	}
	fwrite(b, l, 1, stdout);
    }
}

/*
 * Asks a yes/no question to the user.
 */
int
interface_ask(const char *t)
{
    int c;
    int r = -1;

    printf("%s [yn]\a ", t);
    while(r < 0) {
	fflush(stdout);
	c = read_one_char();
	if(c == 'Y' || c == 'y')
	    r = 1;
	else if(c == 'N' || c == 'n')
	    r = 0;
	else
	    putchar('\a');
    }
    putchar('\n');
    return(r);
}
