/***************************************************************************
 * Speextalk
 * (C) Nicolas George - 2005-07-24
 * This code is distributed under the terms of the GNU GPL.
 * speextalk.c: main program
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <unistd.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <pthread.h>
#include <speex/speex.h>

#include "speextalk.h"

const audio_format_t const audio_formats[] = {
    { &speex_nb_mode, 8000, 1 },
    { &speex_wb_mode, 16000, 1 },
    { &speex_uwb_mode, 32000, 1 },
};

const char *remote = NULL;
const char *service = "20977";
const char *capture_device = "default";
const char *playback_device = "default";
const char *comment = NULL;
int max_packet_size = 1444;
int max_packet_frames = 1;
volatile int igain = 100;
int packet_overhead = 20;

int sock;
unsigned char recv_buffer[MAX_PACKET_SIZE];
int l_recv_buffer;
int audio_format_in = 2;
int audio_format_out = 2;
int speex_quality = 8;
int speex_vbr = 1;

volatile int power_out = 0;
volatile int power_in = 0;
volatile int bandwidth_out = 0;
volatile int bandwidth_in = 0;

pthread_mutex_t global_mutex;
volatile int running = 0;

/*
 * Dies printing the system error message.
 */
void
fatal_system(const char *msg)
{
    perror(msg);
    exit(1);
}

/*
 * Updates the mean frame power according the data of the given frame.
 */
void
update_power(volatile int *power, short *f, int fs)
{
    int p = 0, v, i;

    for(i = 0; i < fs; i++) {
	v = f[i];
	if(v < 0)
	    v = -v;
	if(v > p)
	    p = v;
    }
    pthread_mutex_lock(&global_mutex);
    *power = (*power * 15 + p + 8) / 16;
    pthread_mutex_unlock(&global_mutex);
}

/*
 * Updates the bandwidth statistics.
 */
void
update_bandwidth(volatile int *bandwidth, int *acc, unsigned *timestamp,
    int size)
{
    unsigned t;
    struct timeval tv;

    size += packet_overhead;
    gettimeofday(&tv, NULL);
    t = (unsigned)tv.tv_sec;
    if(*timestamp != t) {
	*timestamp = t;
	pthread_mutex_lock(&global_mutex);
	*bandwidth = *acc;
	pthread_mutex_unlock(&global_mutex);
	*acc = 0;
    }
    *acc += size;
}

/*
 * Dies if f is not a valid audio format.
 */
void
check_audio_format(int f, const char *t)
{
    if(f < 1 || f > (int)(sizeof(audio_formats) / sizeof(*audio_formats))) {
	fprintf(stderr, "Invalid %saudio format.\n", t);
	exit(1);
    }
}

/***************************************************************************
 * Network and threads
 ***************************************************************************/

/*
 * Creates the audio loops, runs the interface, kills the audio loops.
 */
static void
start_loops(void)
{
    pthread_t te, td;

    running = 1;
    pthread_mutex_init(&global_mutex, NULL);
    if(pthread_create(&td, NULL, decode_loop, NULL))
	fatal_system("pthread_create");
    if(pthread_create(&te, NULL, encode_loop, NULL))
	fatal_system("pthread_create");
    interface_run();
    pthread_cancel(te);
    pthread_cancel(td);
    pthread_join(te, NULL);
    pthread_join(td, NULL);
}

/*
 * Establishes a passive socket, waits for handshake and starts the main
 * loop.
 */
static void
wait_connection(void)
{
    struct addrinfo *addr, *a;
    struct addrinfo hints = {
	.ai_flags = AI_PASSIVE,
	.ai_family = AF_UNSPEC,
	.ai_socktype = SOCK_DGRAM,
	.ai_protocol = 0,
    };
    int e;
    int s = -1;

    if((e = getaddrinfo(NULL, service, &hints, &addr))) {
	const char *te = gai_strerror(e);
	fprintf(stderr, "getaddrinfo: %s\n", te);
	exit(1);
    }
    for(a = addr; a != NULL; a = a->ai_next) {
	if((s = socket(a->ai_family, a->ai_socktype, a->ai_protocol)) < 0)
	    continue;
	if(bind(s, a->ai_addr, a->ai_addrlen) < 0) {
	    e = errno;
	    close(s);
	    s = -1;
	    errno = e;
	}
	break;
    }
    if(s < 0)
	fatal_system("wait_connection");
    freeaddrinfo(addr);
    sock = s;

    wait_handshake();
    start_loops();
}

/*
 * Connects to a remote socket, starts the handshake and starts the main
 * loop.
 */
static void
connect_remote(const char *remote)
{
    struct addrinfo *addr, *a;
    struct addrinfo hints = {
	.ai_flags = 0,
	.ai_family = AF_UNSPEC,
	.ai_socktype = SOCK_DGRAM,
	.ai_protocol = 0,
    };
    int e;
    int s = -1;

    if((e = getaddrinfo(remote, service, &hints, &addr))) {
	const char *te = gai_strerror(e);
	fprintf(stderr, "getaddrinfo: %s\n", te);
	exit(1);
    }
    for(a = addr; a != NULL; a = a->ai_next) {
	if((s = socket(a->ai_family, a->ai_socktype, a->ai_protocol)) < 0)
	    continue;
	if(connect(s, a->ai_addr, a->ai_addrlen) < 0) {
	    e = errno;
	    close(s);
	    s = -1;
	    errno = e;
	}
	break;
    }
    if(s < 0) {
	perror("connect_remote");
	exit(1);
    }
    freeaddrinfo(addr);
    sock = s;

    start_handshake();
    start_loops();
}

/*
 * Prints the usage help and terminates with the error code r.
 */
static void
usage(int r)
{
    fprintf(r ? stderr : stdout,
	"Usage: speextalk [options] [remote]\n"
	"\t-g gain: input gain in percent (default: 100)\n"
	"\t-c device: capture device\n"
	"\t-p device: playback device\n"
	"\t-s service: service (port) (default: 20977)\n"
	"\t-m mode: audio mode\n"
	"\t-f frames: max frames per packet (default: 1)\n"
	"\t-q quality: encoding quality (default: 8, legal: 0-10)\n"
	"\t-v 0/1: variable bitrate (default: 1)\n"
	"\t-C comment: connection comment (default: your name or login)\n");
    exit(r);
}

/*
 * Parses the command line, inits the interface and runs wait_connection or
 * connect_remote.
 */
int
main(int argc, char **argv)
{
    {
	int o;

	while((o = getopt(argc, argv, "hg:c:p:s:m:f:q:v:C:")) != -1) {
	    switch(o) {
		case 'g':
		    igain = strtol(optarg, NULL, 0);
		    break;
		case 'c':
		    capture_device = optarg;
		    break;
		case 'p':
		    playback_device = optarg;
		    break;
		case 's':
		    service = optarg;
		    break;
		case 'm':
		    audio_format_out = strtol(optarg, NULL, 0);
		    break;
		case 'f':
		    max_packet_frames = strtol(optarg, NULL, 0);
		    break;
		case 'q':
		    speex_quality = strtol(optarg, NULL, 0);
		    break;
		case 'v':
		    speex_vbr = strtol(optarg, NULL, 0);
		    break;
		case 'C':
		    comment = optarg;
		    break;
		case 'h':
		    usage(0);
		default:
		    usage(1);
	    }
	}
	argc -= optind;
	argv += optind;
    }
    interface_init();
    if(argc == 1)
	connect_remote(argv[0]);
    else
	wait_connection();
    return(0);
}
